/*
	Copyright 1985, 1986, 1987, 1988, 1989, 1990, 1991 Chris Lewis
		All Rights Reserved

    See the LICENSE file for a full description of restrictions under which
    this software is provided.

    Function:		Generic Postscript filter (page reversal etc.)
 */

/* If you wish to add new shuffle algorithms, you will have to:
	- update the getopt for the new option letter (see /OPTION/)
	  by adding the new option letter to the getopt string.
	- update the argument parsing to recognize the new option
	  (/OPTIONPARSE/) by adding a case clause.
	- update the "transform" routine to call your shuffle function,
	  selected upon the option letter.
	- add a new transform routine (see the "o_" functions for
	  samples)
   If you write one, please let me know and I'll reintegrate it
   into the original.  Please note: I will assume that any contributions
   are public domain.
 */
#ifndef	lint
static char SCCSid[] =
    "@(#)psxlate.c: 2.19 Copyright 92/01/23 14:22:16 Chris Lewis";
#endif

#ifdef	ALONE
#include <stdio.h>
#include <ctype.h>
#else
#include "defs.h"
#endif

#ifdef	REGSPILL
#define	register
#endif

int	xlate;
int	italic;
int	needd;
#define	BUFFERSIZE	512

int	half = 0;	/*	Controls emitter, 0: whole file, 1: first half,
				2: second half */

/* define if you want to page flip %! documents (eg: some TeX out stuff)
   not recommended....
 */
#undef	TRYPERBANG

struct range {
    int start;
    int end;
} ranges[100];

int ri;

struct pagedesc {
    unsigned long start;
    unsigned long end;
} pageidx[MAXPAGE], prolog, epilog;

struct pagedesc blankpage = {-1, -1};

char shuffle[100];

#ifdef VMS
char tmp[] = {"SYS$SCRATCH:psxXXXXXX.tmp"};
#else
char tmp[] = {"/tmp/psxXXXXXX"};
#endif /* ifdef VMS */

int pagecnt = 0;
char *progname;
int debug, verbose;
char *mygets();

#ifndef	LIBDIR
#ifdef VMS
#define	LIBDIR	"PSROFF_POSTSCRIPT"
#else
#define	LIBDIR	"/usr/lib/troff2"
#endif /* ifdef VMS */
#endif

char	cmdbuffer[BUFFERSIZE];	/* for psnup stuff */
int	psnupok = 0, psnupwanted = 0;
int	reversed = 0, pages = 2, startspot = 0;

char *magicstr;

FILE *tempfile;

main(argc, argv)
int argc; char **argv; {
    extern int optind;
    extern char *optarg;
    char *buffer;
    int c;
    char *si = shuffle;
    long offset, oldoffset;

#ifdef VMS
    if(!shell$is_shell())
      shell_mung(&argc,&argv,0,"n-p-s-");
#endif /* ifdef VMS */


    progname = argv[0];


    while ((c = getopt(argc, argv, "i2drbcvBEn:p:s:")) != EOF)	/* OPTION */
	switch(c) {
	    /* OPTIONPARSE */
	    case '2':
	    case 'r':
	    case 'b':
	    case 'c':
		*si++ = c;
		if (c == 'r')
		    reversed = 1;
		break;
	    case 'i':
		italic = 1;
		break;
	    case 'p':
		psnupwanted = 1;
		pages = atoi(optarg);
		if (pages != 2 && pages != 4 && pages != 8 && pages != 16) {
		    fprintf(stderr, "%s: -p must be 2, 4, 8 or 16\n",
			progname);
		    exit(1);
		}
		break;
	    case 's':
		psnupwanted = 1;
		startspot = atoi(optarg);
		break;
	    case 'E': half++;
	    case 'B': half++;
		break;
	    case 'd': debug = 1; break;
	    case 'v': verbose = 1; break;
	    case 'n':
		readranges(optarg);
		break;
	    default:
		usage();
		exit(1);
	}

    if (psnupwanted) {
	psnupok = 1;
	if (startspot < 0 || startspot >= pages) {
	    fprintf(stderr, "%s: -s (%d) must be 0..%d (-p%d - 1)\n",
		progname, startspot, pages - 1, pages);
	    psnupok = 0;
	}
#ifdef VMS
	sprintf(cmdbuffer, "%s:nup.pro", LIBDIR);
#else
	sprintf(cmdbuffer, "%s/nup.pro", LIBDIR);
#endif /* ifdef VMS */
	if (access(cmdbuffer, 4) != 0) {
	    fprintf(stderr, "%s: can't do n-up (no nup.pro at %s)\n",
		progname, cmdbuffer);
	    psnupok = 0;
	}
#ifdef VMS
	sprintf(cmdbuffer, "%s:nup.epi", LIBDIR);
#else
	sprintf(cmdbuffer, "%s/nup.epi", LIBDIR);
#endif /* ifdef VMS */
	if (access(cmdbuffer, 4) != 0) {
	    fprintf(stderr, "%s: can't do n-up (no nup.epi at %s)\n",
		progname, cmdbuffer);
	    psnupok = 0;
	}
	if (!psnupok)
	    exit(1);
    }

    if (debug) {
	fprintf(stderr, "startspot: %d\n", startspot);
	fprintf(stderr, "pages: %d\n", pages);
	fprintf(stderr, "reversed: %d\n", reversed);
	fprintf(stderr, "psnupwanted: %d\n", psnupwanted);
	fprintf(stderr, "xlate: %s\n", shuffle);
	for (c = 0; c < ri; c++)
	    fprintf(stderr, "RANGE %d: %d-%d\n", c, ranges[c].start, ranges[c].end);
    }

    mktemp(tmp);
    tempfile =  fopen(tmp, "w");
    if (!tempfile) {
	fprintf(stderr, "%s: couldn't open temp file %s\n",
	    progname, tmp);
	exit(1);
    }
    if (buffer = mygets(stdin, tempfile, &offset)) {
	if (strncmp(buffer, "%!PS-Adobe-", 11) == 0) {
	    if (verbose || debug)
		fprintf(stderr, "DEBUG: conformant Postscript file %s\n", buffer);

	    scan(stdin, tempfile, buffer);
	    stripbyrange();
	    transform();
	    emit(tmp);
	} else if (strncmp(buffer, "%!", 2) == 0) {
	    if (verbose || debug)
		fprintf(stderr, "DEBUG: nonconformant Postscript file %s\n", buffer);
	    fputs(buffer, stdout);
	    while ((c = fread(buffer, 1, sizeof(buffer), stdin)) > 0)
		fwrite(buffer, 1, c, stdout);
	} else {
	    if (verbose || debug)
		fprintf(stderr, "DEBUG: non-Postscript file %s\n", buffer);
	    asc2ps(buffer);
	    fclose(tempfile);
	    stripbyrange();
	    transform();
	    emit(tmp);
	}
    }
    cleanup();
    exit(0);
}

/*	States: */
#define	PROLOG	0
#define	PAGES	1
#define	EPILOG	2

/*	What we've just seen */
#define	S_NONE	0
#define	S_EOF	1
#define	S_PAGE	2
#define	S_EPI	3

scan(in, out, buffer)
FILE *in;
FILE *out;
char *buffer; {
    extern char *malloc();
    long offset;
    int state = PROLOG;
    int seen;
    int nest = 0;

    magicstr = malloc(strlen(buffer) + 1);
    allchk(magicstr);
    strcpy(magicstr, buffer);

    while(1) {
	buffer = mygets(in, out, &offset);

	if (!prolog.start)
	    prolog.start = offset;

	if (!buffer)
	    seen = S_EOF;
	else if (buffer[0] == '%' && buffer[1] == '%') {
	    if (!nest && strncmp(buffer, "%%Page:", 7) == 0)
		seen = S_PAGE;
	    else if (!nest && isspace(buffer[9]) &&
		strncmp(buffer, "%%Trailer", 9) == 0)
		seen = S_EPI;
	    else {

		if (strncmp(buffer, "%%BeginDocument:", 16) == 0)
		    nest++;
		else if (strncmp(buffer, "%%BeginData:", 12) == 0)
		    nest++;
		else if (isspace(buffer[9]) &&
		    strncmp(buffer, "%%EndData", 9) == 0)
		    nest--;
		else if (isspace(buffer[13]) &&
		    strncmp(buffer, "%%EndDocument", 13) == 0)
		    nest--;

		seen = S_NONE;

	    }
	} else
	    seen = S_NONE;
	switch(seen) {
	    case S_NONE:
		continue;
	    case S_EOF:
		switch(state) {
		    case PROLOG:
			prolog.end = offset - 1;
			break;
		    case PAGES:
			pageidx[pagecnt].end = offset - 1;
			pagecnt++;
			break;
		    case EPILOG:
			epilog.end = offset - 1;
			break;
		}
		fclose(out);
		if (psnupwanted)
		    append(pages);
		goto endit;
	    case S_PAGE:
		switch(state) {
		    case PROLOG:
			prolog.end = offset - 1;
			state = PAGES;
			pagecnt = 0;
			if (psnupwanted)
			    while(startspot) {
				pageidx[pagecnt++].start = -1;
				startspot--;
			}
			break;
		    case PAGES:
			pageidx[pagecnt].end = offset - 1;
			pagecnt++;
			break;
		}
		pageidx[pagecnt].start = offset;
		break;
	    case S_EPI:
		switch(state) {
		    case PROLOG:
			prolog.end = offset - 1;
			break;
		    case PAGES:
			pageidx[pagecnt].end = offset - 1;
			pagecnt++;
			break;
		}
		epilog.start = offset;
		state = EPILOG;
		break;
	}
    }
endit:
    if (nest) {
	fprintf(stderr, "%s: misnested BeginDocument/EndDocument\n",
	    progname);
    }
}

cleanup() {
    if (!debug)
#ifdef VMS
        delete(tmp);
#else
	unlink(tmp);
#endif /* ifdef VMS */
}

emit(fname)
char *fname; {
    register int i, start, end;
    FILE *file;
    if ((file = fopen(fname, "r")) == NULL) {
	fprintf(stderr, "Panic %s\n", fname);
	exit(1);
    }

    if (magicstr)
	fputs(magicstr, stdout);

    if (psnupwanted) {
#ifdef VMS
	sprintf(cmdbuffer, "%s:nup.pro", LIBDIR);
#else
	sprintf(cmdbuffer, "%s/nup.pro", LIBDIR);
#endif /* ifdef VMS */
	emitnup(cmdbuffer, 1);
    }

    emitsect(file, &prolog);

    start = 0;
    end = pagecnt - 1;

    if (half == 1)
	end = end / 2;
    else if (half == 2)
	start = end / 2 + 1;

    for (i = start; i <= end; i++)
	emitsect(file, &pageidx[i]);

    emitsect(file, &epilog);

    if (psnupwanted) {
#ifdef VMS
	sprintf(cmdbuffer, "%s:nup.epi", LIBDIR);
#else
	sprintf(cmdbuffer, "%s/nup.epi", LIBDIR);
#endif /* ifdef VMS */
	emitnup(cmdbuffer, 0);
    }

    if (needd)
	putchar('\004');
}

emitsect(file, sect)
FILE *file;
struct pagedesc *sect; {
    long start, end, cnt;

    if (sect->start == 0 && sect->end == 0)
	return;

    if (sect->start == -1) {
	printf("%%%%Page: ? x\n");
	printf("showpage\n");
	return;
    }

    start = sect->start;
    end = sect->end;
    cnt = end - start + 1;

    if (debug)
	fprintf(stderr, "Emitting a section: %d/%d(%d)\n", start, end, cnt);

    fseek(file, start, 0);
    while(cnt--)
	putchar(getc(file));
}

usage() {
    fprintf(stderr,
	"Usage: %s [-b2rc*] [-pn] [-sn] [-B|-E] [-n...] < file > file\n",
	progname);
}

dumppg() {
    register int i;
    if (!debug)
	return;
    fprintf(stderr, "Prolog: %d %d\n", prolog.start, prolog.end);
    for (i = 0; i < pagecnt; i++)
	fprintf(stderr, "%d: %d %d\n", i, pageidx[i].start, pageidx[i].end);
    fprintf(stderr, "Epilog: %d %d\n", epilog.start, epilog.end);
}


transform() {
    register char *p = shuffle;
    dumppg();
    while(*p) {
	switch(*p++) {
	    case '2':
		o_duplex();
		break;
	    case 'r':
		o_reverse();
		break;
	    case 'b':
		o_book();
		break;
	    case 'p':
		o_dupswap();
		break;
	}
    }
    dumppg();
}

/*	Make the number of pages a multiple of n,
	by appending blank pages
 */
append(n)
int n; {
    while(pagecnt % n)
	pageidx[pagecnt++] = blankpage;
}

/*	1 2 3 4 5 6 7 8 -> 1 3 5 7 ... 2 4 6 8 */
o_duplex() {
    register int i, mid;
    struct pagedesc t;
    struct pagedesc *p;

    append(2);

    p = (struct pagedesc *) malloc(sizeof(struct pagedesc) * pagecnt);
    allchk(p);

    mid = pagecnt / 2;

    for (i = 0; i < mid; i++) {
	p[i] = pageidx[i * 2];
	p[i+mid] = pageidx[i * 2 + 1];
    }

    memcpy((char *) pageidx, p, sizeof(struct pagedesc) * pagecnt);
    free(p);
}

o_dupswap() {
    register int i, j;
    struct pagedesc t;

    append(2);

    i = 0;
    j = pagecnt - 1;

    while(i < j) {
	t = pageidx[i];
	pageidx[i] = pageidx[j - 1];
	pageidx[j - 1] = t;
	t = pageidx[i+1];
	pageidx[i+1] = pageidx[j];
	pageidx[j] = t;
	i+=2;
	j-=2;
    }
}

o_reverse() {
    register int i, j;
    struct pagedesc t;

    i = 0;
    j = pagecnt - 1;

    while(i < j) {
	t = pageidx[i];
	pageidx[i] = pageidx[j];
	pageidx[j] = t;
	i++;j--;
    }
}

o_book() {
    struct pagedesc *p;
    int i, mid, end;
    /*  1 2 3 4 5 6 7 8 ... n -> n 1 n-2 3 n-4 5        2 n-1 4 n-3 ... */

    append(4);

    p = (struct pagedesc *) malloc(sizeof(struct pagedesc) * pagecnt);
    allchk(p);
    mid = pagecnt / 2;
    for (i = 0, end = pagecnt - 1; i < mid; i += 2, end -= 2) {
	p[i] = pageidx[end];
	p[i+1] = pageidx[i];
    }
    for (i = mid, end = pagecnt - 1; i < pagecnt; i += 2, end -= 2) {
	p[i] = pageidx[i - mid + 1];
	p[i+1] = pageidx[end-1];
    }
    memcpy((char *) pageidx, p, sizeof(struct pagedesc) * pagecnt);
    free(p);

}

/*	Read one line of file in, writing as we go to out,
	storing the file offset of the beginning of the file
	thru *count;
 */
#define	CHUNK	512
char *
mygets(in, out, count)
FILE *in, *out; long *count; {
    static char *buffer = NULL;
    static long buflimit;
    static long curcount = 0;

    register char *p;
    register long curlength;
    register int ch;
    extern char *realloc();

    if (count != (long *) NULL)
	*count = curcount;

    if (buffer == NULL) {
	buffer = malloc(4096);
	allchk(buffer);
	buflimit = 4096;
    }

    p = buffer;
    while((ch = getc(in)) != EOF) {
	putc(ch, out);
	curcount++;
	if (p - buffer > buflimit-3) {
	    int offset = p - buffer;
	    buflimit += CHUNK;
	    buffer = realloc(buffer, buflimit);
	    allchk(buffer);
	    p = buffer + offset;
	}
	*p++ = ch;
	if (ch == '\n')
	    break;
    }

    *p = '\0';

    if (p == buffer)	/* EOF */
	return(NULL);
    else if (buffer[0] == '\004' && !buffer[1]) {
	needd = 1;
	return(NULL);
    } else
	return(buffer);
}

rangecmp(a, b)
struct range *a, *b; {
    if (a->start < b->start)
	return(-1);
    else if (a->start > b->start)
	return(1);
    else
	return(0);
}

readranges(s)
char *s; {
    extern char *getrange();
    register int i;

    while(s && *s)
	s = getrange(s);

    /*	Internally, pages are 0 origin */
    for (i = 0; i < ri; i++) {
	ranges[i].start--;
	ranges[i].end--;
    }

    qsort(ranges, ri, sizeof(struct range), rangecmp);
}

char *
getrange(s)
char *s; {
    register char *p = s;
    int tmp = 0;
    ranges[ri].end = -1;

    if (*s == '-')
	tmp = 1;
    else {
	while (isdigit(*s))
	    tmp = 10 * tmp + (*s++ - '0');
	if (tmp <= 0) {
	    fprintf(stderr, "%s: bad number %s\n", progname, p);
	    exit(1);
	}
    }

    ranges[ri].end = ranges[ri].start = tmp;

    if (*s == '-') {
	s++;
	p = s;
	tmp = 0;
	if (*s == ',' || *s == '\0')
	    tmp = 100000;
	else {
	    while(isdigit(*s))
		tmp = 10 * tmp + (*s++ - '0');
	    if (tmp <= 0) {
		fprintf(stderr, "%s: bad number %s\n", progname, p);
		exit(1);
	    }
	}
	ranges[ri].end = tmp;
    }

    if (*s == ',')
	s++;
    ri++;
    return(s);
}

stripbyrange() {
    int i, j, prevcnt = pagecnt;

    if (ri == 0)
	return;

    for (i = 0; i < pagecnt; i++) {
	for (j = 0; j < ri; j++)
	    if (i >= ranges[j].start && i <= ranges[j].end)
		break;
	if (j == ri)
	    pageidx[i].start = 0;
    }

    if (debug)
	for (i = 0; i < pagecnt; i++)
	    fprintf(stderr, "SCAN: %d: %d/%d\n", i, pageidx[i].start,
		pageidx[i].end);

    for (i = 0, j = 0, prevcnt = pagecnt; i < prevcnt; i++)
	if (pageidx[i].start != 0)
	    pageidx[j++] = pageidx[i];
	else
	    pagecnt--;

    if (debug)
	fprintf(stderr, "%d pages deleted\n", prevcnt - pagecnt);
}

emitnup(file, flag)
char *file;
int flag; {
    FILE *f;

    if ((f = fopen(file, "r")) == NULL) {
	fprintf(stderr, "%s: WOT? %s!  It was there before!\n",
	    progname, file);
	exit(1);
    }
    if (flag) {
	printf("/@#@Pages@#@ %d def\n", pages);
	printf("/@#@Start@#@ %d def\n", startspot);
	printf("/@#@Rev@#@ %s def\n", reversed? "true" : "false");
    }

    while(fgets(cmdbuffer, sizeof(cmdbuffer), f))
	fputs(cmdbuffer, stdout);

    fclose(f);
}

allchk(p)
register char *p; {
    if (!p) {
	fprintf(stderr, "%s: [re|m]alloc failed - out of memory\n",
	    progname);
	exit(1);
    }
}
