/*
	Copyright 1985, 1986, 1987, 1988, 1989, 1990, 1991 Chris Lewis
		All Rights Reserved

    See the LICENSE file for a full description of restrictions under which
    this software is provided.

    Function: 	converts ditroff tables to CAT troff.
 */

#ifndef	lint
static char SCCSID[] =
    "@(#)dit2catwid.c 2.10 Copyright 91/10/17 19:56:13 Chris Lewis";
#endif

#include <stdio.h>
#include "../defs.h"
#include <ctype.h>

#ifndef	ASCIIWIDTHS

#ifndef	FONTDIR
#define	FONTDIR	"/usr/lib/font"
#endif

#ifndef	HEADERSIZE
#	ifdef COFF
#		include <aouthdr.h>
#		include <filehdr.h>
#		include <scnhdr.h>
#		include <ldfcn.h>
#		include <reloc.h>
#		define H1 (sizeof (struct filehdr) + sizeof (struct aouthdr))
#		define H2 (sizeof (struct scnhdr) + sizeof (struct reloc))
#		define HEADERSIZE (H1 + H2)
#	else
#		include <a.out.h>
#		define HEADERSIZE (sizeof (struct exec))
#	endif
#endif

#endif

int res = 300;
int unitwid = 10;
int symfont = 0;
int verbose = 0;
int genc = 0;
int headersize = -1;
int ext = 0;
FILE *extensions;

#define	TROFFUWID	6

#ifdef	DEBUG
#undef	DEBUG
#endif

FILE *mustopen();
char *mustmalloc();

extern struct cattab tabN[], tabS[];

char *progname;

main(argc, argv)
int argc; char **argv; {
    extern char *optarg;
    extern int optind;
    char *p;
    int c;

    FILE *in1, *in2, *out;
    char *fin1, *fin2, *fout;

    progname = strrchr(argv[0], '/');
    if (progname)
	progname++;
    else
	progname = argv[0];

    while ((c = getopt(argc, argv, "evsr:u:chH:")) != EOF)
	switch(c) {
	    case 'h':
#ifdef	ASCIIWIDTHS
		printf("ascii\n");
#else
		printf("%d\n", HEADERSIZE);
#endif
		exit(0);
		break;
	    case 'H':
		headersize = atoi(optarg);
		break;
	    case 's':
		symfont = 1;
		break;
	    case 'r':
		res = atoi(optarg);
		break;
	    case 'u':
		unitwid = atoi(optarg);
		break;
	    case 'e':
		ext = 1;
		break;
	    case 'v':
		verbose = 1;
		break;
	    case 'c':
		genc = 1;
		break;
	    default:
		usage();
	}
    if (optind != argc - 3) {
	usage();
    }
    in1 = mustopen(fin1 = argv[optind], "r");
    in2 = mustopen(fin2 = argv[optind+1], "r");
    out = mustopen(fout = argv[optind+2], "w");

    process(in1, fin1);
    if (ext) {
	p = mustmalloc(strlen(argv[optind+2]+15), "extension file");
	strcpy(p, argv[optind+2]);
	strcat(p, ".ext");
	extensions = mustopen(p, "w");
    }
    process(in2, fin2);
    if (ext)
	fclose(extensions);

    sort(tabS);
    sort(tabN);

    if (genc) {
	if (symfont)
	    genccode(tabS, out, fout, "Symbol character set");
	else
	    genccode(tabN, out, fout, "Normal character set");
    } else {
	if (symfont)
	    output(tabS, out, fout, "Symbol character set");
	else
	    output(tabN, out, fout, "Normal character set");
    }
    fclose(in1);
    fclose(in2);
    fclose(out);
    exit(0);
}

genccode(symset, f, file, type)
FILE *f;
struct cattab *symset;
char *type, *file; {
    int i, j;

    fprintf(f, "/* %s */\n", type);
    fprintf(f, "char ftXX[224] = {\n");

    for (i = 0, j = 0; i < 224; i++) {
	/* skip to next valid CAT char */
	while(symset[j].ch_name && !symset[j].ch_name[0])
	    j++;

	/* not the right position yet */
	if (!symset[j].ch_name || i < symset[j].ch_wididx) {
	    gencline(f, i, 0, NULL);
	    continue;
	}

	if (symset[j].ch_wididx != i) {
	    fprintf(stderr, "GFNTTAB: index mismatch: (%d/%d)\n",
		i, j);
	    exit(1);
	}

	if (!(symset[j].ch_info&0x3f)) {
	    printf("INFO: No width for %s (index %d, file %s)\n",
		symset[j].ch_name, symset[j].ch_wididx, file);
	    gencline(f, i, symset[0].ch_info, symset[j].ch_name);
	} else
	    gencline(f, i, symset[j].ch_info, symset[j].ch_name);
	j++;
    }

    fprintf(f, "};\n");
    dumpset(symset, "After C Dump");

    if (symset[j].ch_name) {
	fprintf(stderr, "GFNTTAB: table error (too long: %d left: %s)\n",
	    j, symset[j].ch_name);
	exit(1);
    }
}

gencline(file, index, width, name)
FILE *file; int index, width; char *name; {
    int w = width&077;
    int k = width&0300;
    fprintf(file, "\t%2d", w);
    if (k)
	fprintf(file, "+0%03o", k);
    if (index != 223)
	fprintf(file, ",");
    if (name && *name) {
	if (!k)
	    fprintf(file, "\t");
	fprintf(file, "\t/* %s */\n", name);
    } else
	fprintf(file, "\n");
}

output(symset, f, file, type)
FILE *f;
struct cattab *symset;
char *type, *file; {
    int i, j;

#ifdef	ASCIIWIDTHS
#define	EMITWID(v,f)	fprintf(f, "%d\n", v)
#else
#define	EMITWID(v,f)	fputc(v, f)

    FILE *of = (FILE *) NULL;
    /* If headersize != 0 AND there's a FONTDIR/ftR, output
       the first headersize bytes of it, then the new width table,
       then the trailing bytes of FONTDIR/ftR.  If there's
       no FONTDIR/ftR, we just prepend headersize 0's.
       The reason for this is that some troff's insist on valid
       headers.
    */

    if (headersize == -1)
	headersize = HEADERSIZE;

    if (headersize != 0) {
	char buf[128];
	of = fopen("/usr/lib/font/ftR", "r");
	if (!of) {
	    sprintf(buf, "%s/%s", FONTDIR, "ftR");
	    of = fopen(buf, "r");
	}
	if (of) {
	    
	    for (i = 0; i < headersize; i++)
		fputc(fgetc(of), f);
	    for (i = 0; i < 224; i++)
		fgetc(of);
	    if (feof(of)) {
		fclose(of);
		of = (FILE *) NULL;
	    }
	} else {
	    for (i = 0; i < headersize; i++)
		fputc('\0', f);
	}
    }
#endif

    for (i = 0, j = 0; i < 224; i++) {

	/* skip to next valid CAT char */
	while(symset[j].ch_name && !symset[j].ch_name[0])
	    j++;

	/* not the right position yet */
	if (!symset[j].ch_name || i < symset[j].ch_wididx) {
	    EMITWID('\0', f);
	    continue;
	}

	if (symset[j].ch_wididx != i) {
	    fprintf(stderr, "GFNTTAB: index mismatch: (%d/%d)\n",
		i, j);
	    exit(1);
	}

	if (!(symset[j].ch_info&0x3f)) {
	    printf("INFO: No width for %s (index %d, file %s)\n",
		symset[j].ch_name, symset[j].ch_wididx, file);
	    EMITWID(symset[0].ch_info, f);
	} else
	    EMITWID(symset[j].ch_info, f);
	j++;
    }

#ifndef	ASCIIWIDTHS
    if (of) {
	while((i = fgetc(of)) != EOF)
	    fputc(i, f);
	fclose(of);
    }
#endif

    dumpset(symset, "After CAT dump");

    if (symset[j].ch_name) {
	fprintf(stderr, "GFNTTAB: table error (too long: %d left: %s)\n",
	    j, symset[j].ch_name);
	exit(1);
    }
}

dumpset(symset, str)
struct cattab *symset;
char *str; {
    int i, j;
    if (verbose) {
	printf("\n%s:\n", str);
	for (i = 0; symset[i].ch_name; i++) {
	    if (strlen(symset[i].ch_name) == 0)
		continue;
	    printf(" %4d", symset[i].ch_wididx);
	    printf("	%d+%03o,",
		symset[i].ch_info & 077, symset[i].ch_info & 0300);
	    if (symset[i].ch_name)
		printf ("	/* \"%s\" */\n", symset[i].ch_name);
	    else
		printf ("\n");
	}
    }
}

usage() {
    fprintf(stderr, "usage: %s [-rres] [-uunitwid] normfont symfont output\n",
	progname);
    exit(1);
}

char *
mustmalloc(n)
register int n; {
    register char *p;
    extern char *malloc();
    p = malloc(n);
    if (!p) {
	fprintf(stderr, "%s: cannot alloc %d bytes\n", progname, n);
	exit(1);
    } else
	return(p);
}

FILE *
mustopen(f, mode)
char *f, *mode; {
    FILE *desc;
    if ((desc = fopen(f, mode)) == NULL) {
	fprintf(stderr, "%s: cannot open file %s mode %s\n", progname, f, mode);
	exit(1);
    } else
	return(desc);
}

char *gettoken();

process(f, filename)
FILE *f;
char *filename; {
    char *sv;
    char buffer[512];
    char *tok;
    int commentsok = 1;
    int widval, oldwidval;
    int eval, oldeval;
    while(fgets(buffer, sizeof(buffer), f) && buffer[0] == '#');

    if (feof(f))	/* NULL file */
	return;

    do {
	tok = gettoken(buffer, (char *)NULL);

	if (!tok)
	    continue;
	if (0 == strcmp(tok, "charset"))
	    break;

	if (0 == strcmp(tok, "spacewidth")) {
	    if (!(tok = gettoken(buffer, 1))) {
		fprintf(stderr, "%s: bad spacewidth line in %s\n",
		    progname, filename);
		exit(1);
	    }
	    widval = atoi(tok);
	    widval = ((double) TROFFUWID * TROFFRESOLUTION * widval /
		(unitwid * res)) + .5;
	    update(tabS, " ", widval);
	    update(tabN, " ", widval);
	    update(tabS, "\\|", widval / 2);
	    update(tabN, "\\|", widval / 2);
	    update(tabS, "\\^", widval / 4);
	    update(tabN, "\\^", widval / 4);
	}
    } while(fgets(buffer, sizeof(buffer), f));

    if (strcmp(tok, "charset")) {
	fprintf(stderr, "%s: %s bad format - no charset line\n", progname,
	    filename);
	exit(1);
    }
    while(fgets(buffer, sizeof(buffer), f)) {
	char str[31];
	if (commentsok && buffer[0] == '#')
	    continue;
	if (strlen(buffer) > 1 && buffer[0] == '#' && !isspace(buffer[1]))
	    continue;

	sv = strchr(buffer, '\n');
	if (sv) *sv = '\0';
	
	commentsok = 0;
#ifdef	DEBUG
	printf("buffer: %s\n", buffer);
#endif
	tok = gettoken(buffer, (char *)NULL);
	if (!tok)
	    continue;
	if (strlen(tok) > 30) {
	    if (verbose)
		printf("%s: %s skipping entry %s\n",
		    progname, filename, buffer);
	}
	strcpy(str, tok);
	tok = gettoken(buffer, 1);
	if (!tok) {
	    fprintf(stderr, "%s: %s bad format on line %s\n",
		progname, filename, buffer);
	    exit(1);
	}
	if (tok[0] == '#' && !isspace(tok[1]))
	    continue;
#ifdef	DEBUG
	printf("token: %s\n", tok);
#endif
	if (0 == strcmp(tok, "\"")) {
	    widval = oldwidval;
	    eval = oldeval;
	} else {
	    /* atoi will truncate tok if it contains non-numeric characters.
	       Which is just fine for groff's wid,a,b,c,... format */
	    widval = ((double) TROFFUWID * TROFFRESOLUTION * atoi(tok) /
		(unitwid * res)) + .5;
	    widval = max(widval, 1);
	    if (widval > 63)
		printf("%s: width value too big for %s (%d)\n",
		    progname, str, widval);
	    tok = gettoken(buffer, 1);
	    if (!tok) {
		fprintf(stderr, "%s: %s bad format on line %s\n",
		    progname, filename, buffer);
		exit(1);
	    }
	    widval |= (atoi(tok) << 6);
	    tok = gettoken(buffer, 1);
	    if (!tok) {
		if (verbose)
		    fprintf(stderr, "%s: %s missing emit code on line %s\n",
			progname, filename, buffer);
	    } else
		sscanf(tok, "%o", &eval);
	}
	if (0 == (update(tabS, str, widval) + update(tabN, str, widval))) {
	    if (extensions)
		fprintf(extensions, "%s\t%s\t%04o\n", str,
		    symfont? "S": "N", eval);
	}
	oldeval = eval;
	oldwidval = widval;
    }
}

update(table, str, val)
struct cattab *table;
char *str;
int val; {
    int i;
    int done = 0;

#ifdef	DEBUG
    printf("Inserting char %s, (%03o)\n", str, val);
#endif

    for (i = 0; table[i].ch_name; i++)
	if (str && 0 == strcmp(table[i].ch_name, str)) {
	    if (verbose && table[i].ch_info && table[i].ch_info != val)
		fprintf(stderr, "Duplicate nonequal width entry (%s/%s:%x/%x)\n",
		    table[i].ch_name, str, table[i].ch_info, val);
	    table[i].ch_info = val;
	    done = 1;
	}
    return(done);
}

char *
gettoken(buf, ptr)
char *buf, *ptr; {
    static char *saveptr;
    static char savebuf[512];
    char *bp = savebuf;
    if (!ptr)
	saveptr = buf;
    while(*saveptr && isspace(*saveptr)) saveptr++;
    if (*saveptr)
	while(*saveptr && !isspace(*saveptr))
	    *bp++ = *saveptr++;
    *bp = '\0';
    if (savebuf[0])
	return(savebuf);
    else
	return((char *) NULL);
}

tabcmp(a, b)
struct cattab *a, *b; {
    return(a->ch_wididx - b->ch_wididx);
}

sort(table)
struct cattab *table; {
    register int i;

    for (i = 0; table[i].ch_name; i++);

    qsort(table, i, sizeof(struct cattab), tabcmp);

    dumpset(table, "After Sort");
}
